/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.smx.components.jdbc;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.StringReader;
import java.lang.reflect.Field;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Types;
import java.util.HashMap;
import java.util.Map;
import java.util.ResourceBundle;

import javax.jbi.management.DeploymentException;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.JbiConstants;
import org.apache.servicemix.common.DefaultComponent;
import org.apache.servicemix.common.ServiceUnit;
import org.apache.servicemix.common.endpoints.PollingEndpoint;

// Referenced classes of package org.apache.servicemix.file:
//            FileEndpointType

public class JdbcPollerEndpoint extends PollingEndpoint
    implements JdbcEndpointType
{
	private static final int DEFAULT_REFRESH_RATE = 60000;
	
	// Statement for creation of the keys table: it's database dependent, so
	// allow the user to confgure it
	private static String KEY_QUERY = "SELECT VALUE FROM KEY_VALUES WHERE ID=?";
	private static String KEY_INSERT = "INSERT INTO KEY_VALUES (ID,VALUE) VALUES (?,?)";
	private static String KEY_UPDATE = "UPDATE KEY_VALUES SET VALUE=? WHERE ID=?";
	
	// Supported databases (the key table creation statement may be database dependent)
	private static String MS_SQL_SERVER = "MS_SQL";
	private static String MySQL_SERVER = "MySQL";
	private static String Oracle_SERVER = "Oracle";
	private static String Sybase_SERVER = "Sybase";
	private static String DB2_SERVER = "DB2";
	private static String Postgre_SQL_SERVER = "Postgre_SQL";
	private static String JDBC_SERVER = "JDBC";
	private static String[] supportedDatabases = {MS_SQL_SERVER, MySQL_SERVER, Oracle_SERVER, 
		Sybase_SERVER, DB2_SERVER, Postgre_SQL_SERVER, JDBC_SERVER};

	private static final Log log = LogFactory.getLog(JdbcPollerEndpoint.class);
	    
    // Fully qualified name of a JDBC driver class
    private String driver;
    // JDBC URL of the database to connect to
    private String URL;
    // The username for database access
    private String userName;
    // The password for database access
    private String password;
    // How often to run the SQL statements that generate the messages
    private int refreshRate;
    // The name of the key column. It should be returned from the root statement, and
    // the value of which is used to update the value of stored key. For further explanation
    // refer to next section.
    private String key;
    // The initial value for the stored key. This value is only used once to initialise the
    // key.
    private String initialKeyValue;
    // If this property exists then the output messages will have the column name as an
    // attribute instead of as an element name.
    private String columnNameAsAttribute;
    // If this property exists and the name property is defined for a statement element
    // then the output messages will have that name as an attribute instead of as an element
    // name.
    private String rowNameAsAttribute;
    // If this property exists then the output messages will have the column result value
    // as an attribute instead of as a CDATA section.
    private String valueAsAttribute;
    // The number of rows to be used to generate an outgoing message out of the root
    // statement result set. Typically, <rowsInMessage>1</rowsInMessage> is used to
    // generate an XML message per each row. In this case, the tranction of the root SELECT
    // SQL statement is executed when all the rows previously returned are processed.
    private int rowsInMessage;
    // A unique ID for this database communication point (unique from other database
    // communication points). It is used to uniquely identify the key that is stored inside
    // Rhapsody.
    private String commPointID;
    // The statement element contains one SQL element that will be executed and a
    // number of other statement elements that will each be executed for every row returned
    // by the SQL statement.
    private StatementElement rootStatement;
    // Status of the component: necessary because the method poll is called 
    private boolean active;
    // Identifier of database: some statements may be database dependent
    private String databaseType;
    
    // Current record management
	private String keyDriver;
	private String keyUserName;
	private String keyPassword;
	private String keyURL;
	private String keyCreationStatement;
	
	// Bundle for database dependent statements
    private ResourceBundle statementsBundle;
    
    
    static Map<Integer, String> jdbcTypesNames = null; 
    // -------
    // This method returns the name of a JDBC type.
    // Returns null if jdbcType is not recognized.
    public static String getJdbcTypeName(int jdbcType) {
        
        // Return the JDBC type name
        return (String)jdbcTypesNames.get(new Integer(jdbcType));
    }
    
    public class ConditionKeyNotFoundException extends Exception {
		private static final long serialVersionUID = 1L;

		public ConditionKeyNotFoundException(String message) {
			super(message);
		}
	}
    
    // --------
    
    public JdbcPollerEndpoint()
    {
        init();
    }

    public JdbcPollerEndpoint(ServiceUnit serviceUnit, QName service, String endpoint)
    {
        super(serviceUnit, service, endpoint);
        init();
    }

    public JdbcPollerEndpoint(DefaultComponent component, ServiceEndpoint endpoint)
    {
        super(component, endpoint);
        init();
    }
    
    private void init() {
        driver = null;
        URL = null;
        userName = null;
        password = null;
        refreshRate = DEFAULT_REFRESH_RATE;
        key = null;
        initialKeyValue = null;
        columnNameAsAttribute = null;
        rowNameAsAttribute = null;
        valueAsAttribute = null;
        rowsInMessage = 0;
        rootStatement = null;
        commPointID = null;
        active = false;
        databaseType = null;

    	keyDriver = null;
    	keyUserName = null;
    	keyPassword = null;
    	keyURL = null;
    	keyCreationStatement = null;
    	
    	statementsBundle = null;
    	
    	// Use reflection to populate a map of int values to names
        if (jdbcTypesNames == null) {
        	jdbcTypesNames = new HashMap<Integer, String>();
    
            // Get all field in java.sql.Types
            Field[] fields = java.sql.Types.class.getFields();
            for (int i=0; i<fields.length; i++) {
                try {
                    // Get field name
                    String name = fields[i].getName();
    
                    // Get field value
                    Integer value = (Integer)fields[i].get(null);
    
                    // Add to map
                    jdbcTypesNames.put(value, name);
                } catch (IllegalAccessException e) {
                }
            }
        }
    
    }
    
	@Override
	public synchronized void activate() throws Exception {
		super.activate();

    	// Retrieve database driver
    	log.debug("Loading driver " + driver);
    	Class.forName(driver);
    	
    	if (keyDriver != driver) {
        	log.debug("Loading driver " + keyDriver);
	    	Class.forName(keyDriver);
    	}
    	
    	// Load database dependent statements
    	statementsBundle = ResourceBundle.getBundle("org.spagic.smx.components.jdbc.statements");
    	
    	// Try to update the table for keys: if it already exists, the statement
    	// will have no effect
    	try {
    		// Verify if the key table already exists
    		updateConditionKey(null, null);
	        logger.debug("Table for keys already existing");
	        
	        // Verify if the row containing the condition key exists
	        try {
	        	// Retrieve condition key from our database
	            getConditionKey(commPointID);
		        logger.debug("Row for condition key already existing");
	        } catch (ConditionKeyNotFoundException ex) {
	        	// Insert the row containing the initial key value
	            insertConditionKey(initialKeyValue, commPointID);
		        logger.debug("Row for condition key created");
	        }
    	}
    	catch (Exception ex) {
	        logger.debug("Trying to create table for keys");
	        createConditionKeyTable(initialKeyValue, commPointID);
    	}

    	// Component activated
        active = true;
    }

	@Override
	public synchronized void deactivate() throws Exception {
		super.deactivate();
		active = false;
	}

	/* (non-Javadoc)
	 * @see org.apache.servicemix.common.endpoints.PollingEndpoint#poll()
	 * Method invoked on the scheduled interval timing
	 */
	public void poll()
        throws Exception
    {
      if (!active)
    	  return;
      
	  if(logger.isDebugEnabled())
      {
          logger.debug("Scheduling polling with query " + rootStatement.getSql());
      }
      
      pollDatabase();
    }

    /* (non-Javadoc)
     * @see org.apache.servicemix.common.endpoints.ConsumerEndpoint#validate()
     * Method invoked before starting the component
     */
    public void validate()
        throws DeploymentException
    {
        super.validate();
        
        if(databaseType == null)
        {
            throw new DeploymentException("You must specify a database property");
        }
        
        boolean databaseSupported = false;
        for (int i = 0; i < supportedDatabases.length; i++) {
        	if (supportedDatabases[i].equals(databaseType)) {
        		databaseSupported = true;
        		break;
        	}
        }
        if (!databaseSupported) {
            throw new DeploymentException("Database not supported");
        }

        
        if(driver == null)
        {
            throw new DeploymentException("You must specify a driver property");
        }
        if(URL == null)
        {
            throw new DeploymentException("You must specify a URL property");
        }
        if(password == null)
        {
            throw new DeploymentException("You must specify a password property");
        }        
        if(rootStatement == null)
        {
            throw new DeploymentException("You must specify a rootStatement property");
        }
        if(key == null)
        {
            throw new DeploymentException("You must specify a key property");
        }
        if(initialKeyValue == null)
        {
            throw new DeploymentException("You must specify a initialKeyValue property");
        }
        
        if (keyDriver == null) {
        	keyDriver = driver;
        }
        if (keyUserName == null) {
        	keyUserName = userName;
        }
        if (keyPassword == null) {
        	keyPassword = password;
        }
        if (keyURL == null) {
        	keyURL = URL;
        }
    }

    public String getColumnNameAsAttribute() {
		return columnNameAsAttribute;
	}

	public void setColumnNameAsAttribute(String columnNameAsAttribute) {
		this.columnNameAsAttribute = columnNameAsAttribute;
	}

	public String getDatabaseType() {
		return databaseType;
	}

	public void setDatabaseType(String databaseType) {
		this.databaseType = databaseType;
	}

	public String getDriver() {
		return driver;
	}

	public void setDriver(String driver) {
		this.driver = driver;
	}

	public String getInitialKeyValue() {
		return initialKeyValue;
	}

	public void setInitialKeyValue(String initialKeyValue) {
		this.initialKeyValue = initialKeyValue;
	}

	public String getKey() {
		return key;
	}

	public void setKey(String key) {
		this.key = key;
	}

	public String getPassword() {
		return password;
	}

	public void setPassword(String password) {
		this.password = password;
	}

	public int getRefreshRate() {
		return refreshRate;
	}

	public void setRefreshRate(int refreshRate) {
		this.refreshRate = refreshRate;
	}

	public String getRowNameAsAttribute() {
		return rowNameAsAttribute;
	}

	public void setRowNameAsAttribute(String rowNameAsAttribute) {
		this.rowNameAsAttribute = rowNameAsAttribute;
	}

	public int getRowsInMessage() {
		return rowsInMessage;
	}

	public void setRowsInMessage(int rowsInMessage) {
		this.rowsInMessage = rowsInMessage;
	}

	public String getURL() {
		return URL;
	}

	public void setURL(String url) {
		URL = url;
	}

	public String getUserName() {
		return userName;
	}

	public void setUserName(String userName) {
		this.userName = userName;
	}

	public String getValueAsAttribute() {
		return valueAsAttribute;
	}

	public void setValueAsAttribute(String valueAsAttribute) {
		this.valueAsAttribute = valueAsAttribute;
	}
    
	public String getCommPointID() {
		return commPointID;
	}

	public void setCommPointID(String commPointID) {
		this.commPointID = commPointID;
	}

	
	public StatementElement getRootStatement() {
		return rootStatement;
	}

	public void setRootStatement(StatementElement rootStatement) {
		this.rootStatement = rootStatement;
	}

	public String getKeyCreationStatement() {
        if (keyCreationStatement == null) {
        	keyCreationStatement = statementsBundle.getString(databaseType + "_KEY_CREATE");
        }
        return keyCreationStatement;
	}

	public void setKeyCreationStatement(String keyCreationStatement) {
		this.keyCreationStatement = keyCreationStatement;
	}

	/**
	 * Method that performs the database polling
	 */
	protected synchronized void pollDatabase()
    {
    	log.debug("Entering pollDatabase");
    	
    	Connection conn = null;
    	
    	try {
        	conn = openConnection();
            executeStatement(conn, rootStatement);
    	}
    	catch (Exception ex) {
    		log.error(ex);
    	}
    	finally {
    		try {
    			if (conn != null) {
        			closeConnection(conn);
    			}
    		} catch (Exception ex) {
        		log.error(ex);
    		}
    	}
    	log.debug("Exiting pollDatabase");
    }
    
    /**
     * Opens the connection used for polling
     * @return Database connection
     * @throws ClassNotFoundException
     * @throws SQLException
     */
    protected Connection openConnection() throws ClassNotFoundException, SQLException {
    	// Open database connection
    	log.debug("Using URL " + URL);
    	Connection conn = DriverManager.getConnection(URL, userName, password);
    	
    	return conn;
    }
    
    /**
     * Closes the database connection
     * @param conn
     * @throws SQLException
     */
    protected void closeConnection(Connection conn) throws SQLException
    {
    	conn.close();
    }
    
    protected String handleBinary(Object value){
    	
    		/*
    		BufferedOutputStream os;
    	
    		ByteArrayOutputStream baos = new ByteArrayOutputStream();
    		os = new BufferedOutputStream(baos);
    		os.write(aBlob.getBytes(0, (int) aBlob.length()), 0, (int) aBlob
    				.length());
    		os.flush();
    		os.close();
    		
    		baos.toByteArray();
    		*/
    		if (value != null){
    			return new String(Base64.encodeBase64((byte[])value));
    			
    		}else{
    			logger.debug("BinaryValue is Null returning empty string");
    			return "";
    		}
    	
    }
    
    
    protected String handleBlob(Blob aBlob){
    	try{
    		logger.debug("Handling BLOB");
    		
    		byte[] returnBytes;
    		ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
    		logger.debug("Calling Blog get Binary Stream");
    		InputStream inputStream = aBlob.getBinaryStream();

    		int inByte;
    		logger.debug("Writing BLOB TO Temporary Array");
    		while ((inByte = inputStream.read()) != -1){
    			byteArrayOutputStream.write(inByte);
    		}
    		
    		
    		byteArrayOutputStream.flush();
    		returnBytes = byteArrayOutputStream.toByteArray();
    		logger.debug("BLOB Readed in bytearray of length["+returnBytes.length+"]");
    		byteArrayOutputStream.close();
    		return handleBinary(returnBytes);
    		
    		
    		/*
    		BufferedOutputStream os;
	
    		
    		ByteArrayOutputStream baos = new ByteArrayOutputStream();
    		os = new BufferedOutputStream(baos);
    		os.write(aBlob.getBytes(0, (int) aBlob.length()), 0, (int) aBlob
				.length());
    		
    		
    		byte[] byteArr = baos.toByteArray();
    		logger.debug("BLOB Readed in bytearray of length["+byteArr.length+"]");
    		os.flush();
    		os.close();
    		*/
    	}catch (Exception e) {
    		logger.error(e);
    		return "null";
		}
	
    }
    /**
     * Execute the root statement
     * @param conn Database connection
     * @param currentStatement Root statement
     * @throws Exception
     */
    protected void executeStatement(Connection conn, StatementElement currentStatement) throws Exception
    {
    	PreparedStatement stmt = null;
    	ResultSet rs = null;
        String conditionKey = null;
		String keyValue = null;
		
		log.debug("Executing statement " + currentStatement.getSql());
    	try {
        	stmt = conn.prepareStatement(currentStatement.getSql());
        	// Retrieve condition key from our database
            conditionKey = getConditionKey(commPointID);
            // Use it for data retrieval
        	stmt.setString(1, conditionKey);
        	
        	rs = stmt.executeQuery();

        	// Retrieve metadata to get the column names
        	ResultSetMetaData metadata = rs.getMetaData();
        	
        	// Retrieve result column names
        	int columnCount = metadata.getColumnCount();
        	String[] columnNames = new String[columnCount];
        	String[] resultValues = new String[columnCount];
        	for (int i = 0; i < columnCount; i++) {
        		columnNames[i] = metadata.getColumnName(i + 1);
        	}
        	
        	// Fetch all rows
        	int messageCount = 0;
        	StringBuffer collectionMessageString = new StringBuffer();
        	while (rs.next()) {
        		keyValue = rs.getString(key);
        		
            	for (int i = 0; i < columnCount; i++) {
            		
            		logger.debug("Analyze type of column("+ (i+1) +")");
            		logger.debug(" Column ["+(i+1)+"] has type ["+getJdbcTypeName(metadata.getColumnType(i+1))+"]");
            		
            		if (metadata.getColumnType(i+1) == Types.BINARY){
            			
            			logger.debug("Treat column("+ (i+1) +") as BINARY");
            			resultValues[i] = handleBinary(rs.getObject(i + 1));
            			
            		} else if (metadata.getColumnType(i+1) == Types.BLOB){
            			
            			logger.debug("Treat column("+ (i+1) +") as BLOB");
            			resultValues[i] = handleBlob(rs.getBlob(i + 1));
            		
            		} else {
            		
            			logger.debug("Treat column("+ (i+1) +") as normal get Column using rs.getString");
            			resultValues[i] = rs.getString(i + 1);
            		
            		}
            	}
            	
            	// Begin XML
            	buildStartMessageString(currentStatement.getName(), columnNames, resultValues, collectionMessageString);
            	// Produce XML body
            	buildBodyMessageString(currentStatement.getName(), columnNames, resultValues, collectionMessageString);
            	
            	// Execute the subqueries
            	StatementElement[] subQueries = currentStatement.getSubQueries();
            	if (subQueries != null) {
                	for (int i = 0; i < subQueries.length; i++) {
                		StringBuffer subStatementResult = executeSubStatement(conn, subQueries[i], columnNames, resultValues);
                		if (subStatementResult != null) {
                			collectionMessageString.append(subStatementResult);
                		}
                	}
            	}
            	
            	// Close XML
            	buildEndMessageString(currentStatement.getName(), columnNames, resultValues, collectionMessageString);
            	messageCount++;
            	
            	// Check if we collect enough messages to send
            	if (rowsInMessage == messageCount) {
                	// Send message to NMR
                	sendMessage(collectionMessageString);
                	// Restart collecting messages            		
            		messageCount = 0;
            		collectionMessageString = new StringBuffer();
            	}
    		}
        	
        	// Send message to NMR
        	if (collectionMessageString.length() != 0) {
	        	sendMessage(collectionMessageString);
        	}
    	}
    	finally {
        	try {
				if (rs != null) {
					rs.close();
				}
			} finally {
	        	try {
					if (stmt != null) {
						stmt.close();
					}
				} finally {
		        	// Update the condition key for the next iteration
		        	if (keyValue != null) {
		            	updateConditionKey(commPointID, keyValue);
		        	}
				}
			}
    	}
    	
    }

    /**
     * Executes a substatement
     * @param conn Database connection
     * @param currentStatement Substatement
     * @param parentColumnNames Parent statement column names
     * @param parentResultValues Parent statement result values
     * @return StringBuffer containing the XML message of the subquery
     * @throws Exception
     */
    protected StringBuffer executeSubStatement(Connection conn, StatementElement currentStatement, String[] parentColumnNames, String[] parentResultValues) throws Exception
    {
    	PreparedStatement stmt = null;
    	ResultSet rs = null;
		
		log.debug("Preparing statement " + currentStatement.getSql());
    	try {
    		// The statement can contain references to the values of the parent query
    		String originalStatement = currentStatement.getSql();
    		String modifiedStatement = originalStatement;
    		
    		while (modifiedStatement.indexOf('@') != -1) {
    			int atIndex = modifiedStatement.indexOf('@');
    			int endIndex = modifiedStatement.indexOf(' ', atIndex);
    			if (endIndex == -1) {
    				endIndex = modifiedStatement.length();
    			}
    			
    			// Retrieve the parameter name without the @
    			String parameterName = modifiedStatement.substring(atIndex + 1, endIndex);
    			String parameterValue = null;
    			// Evaluate parameter value
    			for (int i = 0; i < parentColumnNames.length; i++) {
    				if (parameterName.equalsIgnoreCase(parentColumnNames[i])) {
    					parameterValue = parentResultValues[i];
    					break;
    				}
    			}
    			
    			if (parameterValue != null) {
    				modifiedStatement = modifiedStatement.replaceFirst("@" + parameterName, "'" + parameterValue + "'");
    			} else {
    				log.error("Unable to retrieve parameter value for parameter " + parameterName);
    				return null;
    			}
    		}
    		
    		log.debug("Executing statement " + modifiedStatement);
        	stmt = conn.prepareStatement(modifiedStatement);
        	rs = stmt.executeQuery();

        	ResultSetMetaData metadata = rs.getMetaData();
        	
        	// Retrieve result column names
        	int columnCount = metadata.getColumnCount();
        	String[] columnNames = new String[columnCount];
        	String[] resultValues = new String[columnCount];
        	for (int i = 0; i < columnCount; i++) {
        		columnNames[i] = metadata.getColumnName(i + 1);
        	}
        	
        	// Fetch all rows
        	StringBuffer childRowsMessageString = new StringBuffer();
        	while (rs.next()) {
            	for (int i = 0; i < columnCount; i++) {
            		resultValues[i] = rs.getString(i + 1);
            	}
            	
            	// Begin XML
            	buildStartMessageString(currentStatement.getName(), columnNames, resultValues, childRowsMessageString);
            	// Produce XML body
            	buildBodyMessageString(currentStatement.getName(), columnNames, resultValues, childRowsMessageString);
            	
            	// Execute the subqueries
            	StatementElement[] subQueries = currentStatement.getSubQueries();
            	if (subQueries != null) {
                	for (int i = 0; i < subQueries.length; i++) {
                		StringBuffer subMessage = executeSubStatement(conn, subQueries[i], columnNames, resultValues);
                		if (subMessage != null) {
                			childRowsMessageString.append(subMessage);
                		}
                	}
            	}
            	
            	// Close XML
            	buildEndMessageString(currentStatement.getName(), columnNames, resultValues, childRowsMessageString);
    		}
        	return childRowsMessageString;
    	}
    	finally {
        	try {
				if (rs != null) {
					rs.close();
				}
			} finally {
	        	if (stmt != null) {
	        		stmt.close();
	    		}
			}
    	}
    }    
    
    /**
     * Insert the first condition key to filter the data to select
     * @param initialKeyValue Parameter defined in component configuration
     * @param commPointID Unique ID for key retrieval
     * @throws Exception
     */
    protected void insertConditionKey(String initialKeyValue, String commPointID) throws Exception
    {
    	Connection conn = DriverManager.getConnection(keyURL, keyUserName, keyPassword);
    	PreparedStatement insertStmt = null;
    	
    	try {
    		// The row doesnt exists: insert it
        	insertStmt = conn.prepareStatement(KEY_INSERT);
        	insertStmt.setString(1, commPointID);
        	insertStmt.setString(2, initialKeyValue);
        	insertStmt.executeUpdate();
    	}
    	finally {
			try {
				if (insertStmt != null) {
					insertStmt.close();
				}
			} finally {
				if (conn != null) {
					conn.close();
				}
			}
		}
    }

    /**
     * Create the condition key table, if it doesn't exist, and insert the 
     * first condition row with the initial key values.
     * @param initialKeyValue Parameter defined in component configuration
     * @param commPointID Unique ID for key retrieval
     * @throws Exception
     */
    protected void createConditionKeyTable(String initialKeyValue, String commPointID) throws Exception
    {
		Connection conn = DriverManager.getConnection(keyURL, keyUserName, keyPassword);
		PreparedStatement stmt = null;
		try {
			String keyCreationStatement = getKeyCreationStatement();
			logger.debug("Key Creation statement[" + keyCreationStatement + "]");
			stmt = conn.prepareStatement(keyCreationStatement);

			try {
				stmt.executeUpdate();
				logger.debug("Table for keys created");

				// Insert the row containing the initial key value
				insertConditionKey(initialKeyValue, commPointID);
				logger.debug("Row for condition key created");
			} catch (Exception ex1) {
				logger.error("Unable to create table for keys", ex1);
				// Throw again the exception to avoid starting correctly the
				// component
				throw ex1;
			}
		} finally {
			try {
				if (stmt != null) {
					stmt.close();
				}
			} finally {
				if (conn != null) {
					conn.close();
				}
			}
		}
	}

    /**
     * Retrieves the condition key to filter the data to select
     * @param commPointID Unique ID for key retrieval
     * @return The condition key
     * @throws Exception
     */
    protected String getConditionKey(String commPointID) throws Exception
    {
    	String result = null;
    	
    	Connection conn = DriverManager.getConnection(keyURL, keyUserName, keyPassword);
    	PreparedStatement selectStmt = null;
    	ResultSet rs = null;
    	
    	try {
        	selectStmt = conn.prepareStatement(KEY_QUERY);
        	selectStmt.setString(1, commPointID);
        	rs = selectStmt.executeQuery();
        	if (rs.next()) {
        		// The row already exists: retrieve current value
        		result = rs.getString(1);
        	} else {
        		// The row doesnt exists: this should not happen NEVER
        		throw new ConditionKeyNotFoundException("Condition key not found");
        	}
    	}
    	finally {
			try {
				if (rs != null) {
					rs.close();
				}
			} finally {
				try {
					if (selectStmt != null) {
						selectStmt.close();
					}
				} finally {
					if (conn != null) {
						conn.close();
					}
				}
			}
		}
    	
    	return result;
    }
    
    /**
	 * Updates the current condition key according to the previous selected data
	 * 
	 * @param commPointID
	 *            Unique ID for key retrieval
	 * @param currentKeyValue
	 *            Last key used
	 * @throws Exception
	 */
    protected void updateConditionKey(String commPointID, String currentKeyValue) throws Exception
    {
    	Connection conn = DriverManager.getConnection(keyURL, keyUserName, keyPassword);
    	PreparedStatement stmt = null;
    	
    	try {
	    	stmt = conn.prepareStatement(KEY_UPDATE);
	    	stmt.setString(1, currentKeyValue);
	    	stmt.setString(2, commPointID);
	    	stmt.executeUpdate();
    	}
    	finally {
        	try {
				if (stmt != null) {
					stmt.close();
				}
			} finally {
	        	if (conn != null) {
	    			conn.close();
	    		}
			}
    	}
    }

    /**
     * Build the message root: for example
     * 	<main>
     * or
     * 	<row name="main">
     * 
     * @param statementName Name of the statement
     * @param columnNames Names for the result columns
     * @param columnValues Values of the result columns
     * @param result StringBuffer of the resulting XML
     * @return StringBuffer of the resulting XML
     */
    protected StringBuffer buildStartMessageString(final String statementName, final String[] columnNames, final String[] columnValues, StringBuffer result)
    {
    	if (rowNameAsAttribute != null) {
        	result.append("<row name=\"");
        	result.append(statementName);
        	result.append("\">");
    	} else {
        	result.append("<");
        	result.append(statementName);
        	result.append(">");
    	}
    	    	
    	return result;
    }

    /**
     * Build the message body: for example
     * 	<id>20234</id>
     * 	<date>2002-1-2 10:32:25</date>
     * or
     * 	<column name="id" value="97320"/>
     *  <column name="date" value="2002-1-1 12:12:34"/>

     * @param statementName Name of the statement
     * @param columnNames Names for the result columns
     * @param columnValues Values of the result columns
     * @param result StringBuffer of the resulting XML
     * @return StringBuffer of the resulting XML
     */
    protected StringBuffer buildBodyMessageString(final String statementName, final String[] columnNames, final String[] columnValues, StringBuffer result)
    {
    	// Following there are some samples of XML according to different attributes values
		//    	columnNameAsAttribute = null
		//    	valueAsAttribute = null
		//    	<PIPPO>123</PIPPO><PLUTO>456</PLUTO>
		//
		//    	columnNameAsAttribute != null
		//    	valueAsAttribute = null
		//    	<column name="PIPPO"/>123<column/><column name="PLUTO"/>456<column/>
		//    	
		//    	columnNameAsAttribute = null
		//    	valueAsAttribute != null
		//    	<PIPPO value="123"/><PLUTO value="456"/>
		//    	
		//    	columnNameAsAttribute != null
		//    	valueAsAttribute != null
		//    	<column name="PIPPO" value="123"/><column name="PLUTO" value="456"/>
    	
    	if (columnNameAsAttribute != null) {
        	for (int count = columnNames.length, i = 0; i < count; i++) {
            	result.append("<column name=\"");
            	result.append(columnNames[i]);
            	result.append("\"");
            	
            	if (valueAsAttribute != null) {
                	result.append(" value=\"");
                	result.append(columnValues[i]);
                	result.append("\"/>");
            	} else {
                	result.append(">");
                	result.append(columnValues[i]);
                	result.append("</column>");
            	}
        	}
    	} else {
        	for (int count = columnNames.length, i = 0; i < count; i++) {
            	result.append("<");
            	result.append(columnNames[i]);
            	
            	if (valueAsAttribute != null) {
                	result.append(" value=\"");
                	result.append(columnValues[i]);
                	result.append("\"/>");
            	} else {
                	result.append(">");
                	result.append(columnValues[i]);
                	result.append("</");
                	result.append(columnNames[i]);
                	result.append(">");
            	}
        	}
    	}
    	
    	return result;
    }

    /**
     * Build the message tail: for example
     * 	</main>
     * or
     * 	</row>

     * @param statementName Name of the statement
     * @param columnNames Names for the result columns
     * @param columnValues Values of the result columns
     * @param result StringBuffer of the resulting XML
     * @return StringBuffer of the resulting XML
     */
    protected StringBuffer buildEndMessageString(final String statementName, final String[] columnNames, final String[] columnValues, StringBuffer result)
    {
    	if (rowNameAsAttribute != null) {
        	result.append("</row>");
    	} else {
        	result.append("</");
        	result.append(statementName);
        	result.append(">");
    	}
    	
    	return result;
    }
    
    /**
     * Send a message to the NMR
     * @param messageStringBuffer
     * @throws Exception
     */
    protected void sendMessage(StringBuffer messageStringBuffer)
        throws Exception
    {
    	log.debug("Sending message: " + messageStringBuffer);

    	// Envelope the string in a "message" tag: if rowsInMessage is specified this is
    	// necessary to ensure a proper XML message
    	messageStringBuffer.insert(0, "<message>");
    	messageStringBuffer.append("</message>");

        InOnly exchange = getExchangeFactory().createInOnlyExchange();
        //String key = EndpointSupport.getKey(this);
        //exchange.setProperty(JbiConstants.SENDER_ENDPOINT, key);
        
        exchange.setProperty(JbiConstants.SENDER_ENDPOINT, service.toString());
        exchange.setProperty(JbiConstants.CORRELATION_ID, exchange.getExchangeId());
        
        configureExchangeTarget(exchange);
        javax.jbi.messaging.NormalizedMessage message = exchange.createMessage();
        exchange.setInMessage(message);
        StreamSource stream = new StreamSource(new StringReader(messageStringBuffer.toString()));
        message.setContent(stream);
        send(exchange);
    }

    public String getLocationURI()
    {
    	// We need a unique name: may use this ?
        return rootStatement.getSql();
    }

    public void process(MessageExchange messageexchange)
        throws Exception
    {
    }

}
