/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.smx.components.restart;

import org.spagic.metadb.dbutils.HibernateUtil;
import org.spagic.metadb.model.XmlNamespace;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.MessageExchangeListener;
import org.apache.servicemix.components.util.TransformComponentSupport;
import org.apache.servicemix.id.IdGenerator;
import org.apache.servicemix.jbi.NoInMessageAvailableException;
import org.apache.servicemix.jbi.container.JBIContainer;
import org.apache.servicemix.jbi.event.ContainerAware;
import org.apache.servicemix.jbi.jaxp.SourceTransformer;
import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.XPath;
import org.dom4j.xpath.DefaultXPath;
import org.hibernate.Session;

/**
 * Service Engine to resend a Message Exchange
 * It receives the id of the JournalInstanceService on the database, containing
 * the Message Exchange to resend.
 */
public class RestartComponent extends TransformComponentSupport implements
		MessageExchangeListener, ContainerAware {

	// Single Restart expressions
	private static final String SINGLE_INSTANCE_RESTART_XPATH_EXPRESSION = "//restart:restartSingleInstance";
	
	// Massive Instance Restart expressions
	private static final String MASSIVE_INSTANCE_RESTART_XPATH_EXPRESSION = "//restart:restartMultipleInstances";

	public static final String CORRELATION_ID_PREFIX = "RESTARTED_ID:";
	
	private static final Log log = LogFactory.getLog(RestartComponent.class);

	private JBIContainer container = null;

	// Map of namespaces uri for evaluating XPath expressions
	public static Map<String, String> namespaceURI = null;
	
	// Generator for correlation ids
	private static IdGenerator idGenerator = null;
	
	public void debug(Object o){
		if (log.isDebugEnabled()){
			log.debug(o);
		}
	}

	public void setContainer(JBIContainer container) {
		this.container = container;
	}	

	@Override
	public void init(ComponentContext smxContext) throws JBIException {
		super.init(smxContext);

		try {
			// Temporary code necessary until container injection doesn't work
			Field field = smxContext.getClass().getDeclaredField("container");
			field.setAccessible(true);
			JBIContainer container = (JBIContainer) field.get(smxContext);
			setContainer(container);
			debug("Retrieved JBI container");
			
			// Load XML namespaces
			if (namespaceURI == null) {
				initNameSpaceURI();
				debug("Initialized namespace URI");
			}
			
			// Create unique id generator
			if (idGenerator == null) {
				idGenerator = new IdGenerator(CORRELATION_ID_PREFIX);
			}
		} catch (Exception ex) {
			log.error("Unable to retrieve JBI container", ex);
			throw new JBIException(ex);
		}
	}
	
	/**
	 * Load the namespaces necessary to evaluate the XPath expressions
	 */
	public synchronized static void initNameSpaceURI(){
		Session aSession = null;
		try {
			namespaceURI = new HashMap<String, String>();
			aSession = HibernateUtil.getSessionFactory().openSession();
			List<XmlNamespace> lNs = aSession.createQuery(" from org.spagic.metadb.model.XmlNamespace").list();
			for (XmlNamespace ns : lNs) {
				namespaceURI.put(ns.getPrefix(), ns.getValue());
			}
		}catch(Throwable t){
			t.printStackTrace();
		}finally{
			if (aSession != null) {
				aSession.close();
			}
		}
	}

	public static Map<String, String> getNamespaceURI(){
		return namespaceURI;
	}
	
	/**
	 * Apply an XPath expression to the DOM document.
	 * @param document The DOM document of the message
	 * @return True if the expression select some node on the DOM
	 */
	private boolean checkExpression(Document document, String xPathExpression) {
		// According to the service type to activate (single/massive)
		// the input XML is different
		XPath genericXPathExpression = new DefaultXPath(xPathExpression);
		genericXPathExpression.setNamespaceURIs(getNamespaceURI());
		return (genericXPathExpression.selectSingleNode(document) != null);
	}
	
	/**
	 * Parse the message inside the exchange to get a DOM document.
	 * @param exchange The exchange containing the message
	 * @return The DOM document of the message
	 * @throws Exception
	 */
	private Document retrieveDOMDocument(NormalizedMessage in) throws Exception {
		// Create DOM for content
		SourceTransformer sourceTransformer = new SourceTransformer();
		String xmlMessage = sourceTransformer.contentToString(in);
		Document document = DocumentHelper.parseText(xmlMessage);

		return document;
	}
		
    public NormalizedMessage getInMessage(MessageExchange exchange) throws NoInMessageAvailableException {
    	return super.getInMessage(exchange);
    }


	@Override
	protected boolean transform(MessageExchange exchange, NormalizedMessage in, NormalizedMessage out) throws Exception {
		debug("Received a request to restart one or more exchange");

		if (container == null) {
			log.error("Container not available: impossible restarting message");
			throw new MessagingException("Container not available: impossible restarting message");
		}
		
		try {
			// Parse the message to get a DOM, only once
			Document document = retrieveDOMDocument(in);
			
			// Check if we have to restart a single exchange or a list of exchanges
			if (checkExpression(document, SINGLE_INSTANCE_RESTART_XPATH_EXPRESSION)) {
				SingleRestart singleRestartManager = new SingleRestart(container, namespaceURI, idGenerator, this);
				singleRestartManager.manageSingleRestart(exchange, document, out);
			} else if (checkExpression(document, MASSIVE_INSTANCE_RESTART_XPATH_EXPRESSION)) {
				MultipleInstanceRestart multipleInstanceRestartManager = new MultipleInstanceRestart(container, namespaceURI, idGenerator, this);
				multipleInstanceRestartManager.manageMassiveRestart(exchange, document, out);
			} else {
				MultipleMessageRestart multipleMessageRestartManager = new MultipleMessageRestart(container, namespaceURI, idGenerator, this);
				multipleMessageRestartManager.manageMassiveMessagesRestart(exchange, document, out);
			}

		} catch (Exception e) {
	    	log.error("Fault on restarting activity", e);
	    	// Throw the exception: the super class should send a Fault message or an error,
	    	// according to the MEP
	    	throw e;
		}
		
		return true;
	}

}
