/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.smx.listeners.sync;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.JbiConstants;
import org.apache.servicemix.jbi.NoInMessageAvailableException;
import org.apache.servicemix.jbi.event.ExchangeEvent;
import org.apache.servicemix.jbi.event.ExchangeListener;
import org.apache.servicemix.jbi.jaxp.StringSource;

public class SynchronizerFaultListener implements ExchangeListener {

	public static final String SYNCHRONIZER_SERVICE = "SYNCHRONIZER_SERVICE";
	public static final String SYNCHRONIZER_ENDPOINT = "SYNCHRONIZER_ENDPOINT";

	private static final Log log = LogFactory.getLog(SynchronizerFaultListener.class);
	
	/* (non-Javadoc)
	 * @see org.apache.servicemix.jbi.event.ExchangeListener#exchangeAccepted(org.apache.servicemix.jbi.event.ExchangeEvent)
	 */
	public void exchangeAccepted(ExchangeEvent event) {
	}

	/* (non-Javadoc)
	 * @see org.apache.servicemix.jbi.event.ExchangeListener#exchangeSent(org.apache.servicemix.jbi.event.ExchangeEvent)
	 */
	public void exchangeSent(ExchangeEvent event) {
		MessageExchange exchange = event.getExchange();
		
		// Consider only InOnly error exchanges
		if (exchange.getStatus() != ExchangeStatus.ERROR) {
			return;
		}
		if (isInAndOut(exchange)) {
			return;
		}

		try {
			// Check if this message was generated by the Synchronizer
			NormalizedMessage in = getInMessage(exchange);

			QName syncService = (QName) in.getProperty(SYNCHRONIZER_SERVICE);
			String syncEndpoint = (String) in.getProperty(SYNCHRONIZER_ENDPOINT);
			
			if ((syncService == null) || (syncEndpoint == null)) {
				// The process doesn't use the Synchronizer
				return;
			}
			log.debug("syncService:" + syncService);
			log.debug("syncEndpoint:" + syncEndpoint);
			
			// Retrieve correlation id
			String processCorrelationId = (String) exchange.getProperty(JbiConstants.CORRELATION_ID);
			if (processCorrelationId == null) {
				log.warn("Process Correlation ID not present in exchange: " + exchange);
				return;
			}
			log.debug("processCorrelationId:" + processCorrelationId);
			
			// Retrieve the delivery channel to send a message to the
			// synchronizer
			DeliveryChannel channel = event.getExchangeSourceContext()
					.getDeliveryChannel();
			MessageExchangeFactory factory = channel
					.createExchangeFactoryForService(syncService);	
            MessageExchange outExchange = factory.createInOnlyExchange();

            // Set the properties necessary for monitoring
        	outExchange.setProperty(JbiConstants.SENDER_ENDPOINT, syncService + ":" + syncEndpoint);

        	if (processCorrelationId != null) {
            	outExchange.setProperty(JbiConstants.CORRELATION_ID, processCorrelationId);
        	}
            
            NormalizedMessage out = outExchange.createMessage();
            
            StringBuffer responseBuffer = new StringBuffer();
			responseBuffer.append("<errorMessage><![CDATA[");
			Exception error = exchange.getError();
			responseBuffer.append((error.getMessage() != null) ? error.getMessage() : error.toString());
			responseBuffer.append("]]></errorMessage>");
			
			out.setContent(new StringSource(responseBuffer.toString()));

			// Send the exchange back to the Syncrhonizer, allowing it to send the
			// response to the InOut exchange
            outExchange.setMessage(out, "in");
            outExchange.setStatus(ExchangeStatus.ACTIVE);
            channel.send(outExchange);
		} catch (Exception ex) {
			log.error(ex);
		}
	}

	/**
	 * A helper method which will return true if the exchange is capable of both
	 * In and Out such as InOut, InOptionalOut etc.
	 * 
	 * @param exchange
	 * @return true if the exchange can handle both input and output
	 */
	public boolean isInAndOut(MessageExchange exchange) {
		return exchange instanceof InOut || exchange instanceof InOptionalOut;
	}

	/**
	 * Returns the in message or throws an exception if there is no in message.
	 */
	protected NormalizedMessage getInMessage(MessageExchange exchange)
			throws NoInMessageAvailableException {
		NormalizedMessage message = exchange.getMessage("in");
		if (message == null) {
			throw new NoInMessageAvailableException(exchange);
		}
		return message;
	}

}
